/* ++++++++++
	USB Kitty driver.c
+++++ */

#include <OS.h>
#include <Drivers.h>
#include <KernelExport.h>
#include <Errors.h>
#include <stdlib.h>
#include <stdio.h>
#include <malloc.h>
#include <string.h>
#include <USB.h>
#include <USB_spec.h>

#include "usb_kitty.h"

/* $B%f!<%6Dj5ADj?t(B */
#define VENDOR_ID  0x04c5
#define PRODUCT_ID 0x1010
#define ID         "\033[34musb_kitty:\033[m "
#define MAX_DEVICE_NUM   1

typedef struct mydev mydev;
struct mydev {
  mydev                        *next;
  const usb_device             *dev;
  const usb_configuration_info *conf;
  const usb_interface_info     *intf;
  const usb_endpoint_info      *ept;
  void             *buf;
  area_id          buf_area;
  int              num;
  int              open;
  sem_id           lock;
  sem_id           data_lock;
  sem_id           cmd_complete;
  status_t         cmd_status;
  char             button;
  uint32           actual_len;
  bigtime_t        timestamp;
};

/* $B%0%m!<%P%kJQ?t(B */
int32 api_version = B_CUR_DRIVER_API_VERSION;
static usb_module_info *usb; /* USB$B%b%8%e!<%k$N>pJs$rJ]B8$9$k9=B$BN(B */
static const char *driver_name = "usb_kitty"; /* $B%I%i%$%P$N<BBN$N%U%!%$%kL>(B */
static const char *base_name = "misc/usb_kitty/"; /* devfs$B$K$*$1$k%G%P%$%9$N(B
						     $B%Y!<%9J8;zNs(B
						     /dev/misc/usb_kitty/0, 
						     /dev/misc/usb_kitty/1, ...
						  */
static int device_count = 0; /* $B%G%P%$%9?t(B */
static mydev *device_list = NULL; /* $B%G%P%$%9%j%9%H(B*/
static sem_id device_list_lock = -1; /* $B%G%P%$%9%j%9%H$N(Block$BJQ?t(B */
static char **device_names = NULL; /* $B%G%P%$%9%U%!%$%kL>(B */
static volatile char data[256];


/* $B%W%m%H%?%$%W@k8@(B */
static void 
cb_command_send(void *cookie, uint32 status, void *data, uint32 actual_len);

static status_t
kitty_command_send(mydev *md, uint16 command);

static mydev *
add_device(const usb_device *dev);

static void
remove_device(mydev *md);

static status_t
init_device(mydev *new_dev, const usb_device *usb_dev, const int32 num);

static void
deinit_device(mydev *md);

/* $B%m!<%+%k4X?t(B */
static void 
cb_command_send(void *cookie, uint32 status, void *data, uint32 actual_len)
{
  mydev *md = (mydev *)cookie;
  
  //   DPRINTF((ID "cb_command_send(cookie=%p, status=%x, data=%p, actual_len=%x)\n",
  //	    cookie, status, data, actual_len));
  // queue_interrupt()$B4X?t$d(Bqueue_request()$B40N;;~$K%3!<%k$5$l$k!%(B
  // $B40N;;~$K<B9T$7$?$$=hM}$r=q$/!%(B

  if (md != NULL) {
    md->cmd_status = status;
    md->actual_len = actual_len;
    release_sem(md->cmd_complete);
    release_sem(md->data_lock);
  }
  else {
    delete_sem(md->cmd_complete);
    delete_sem(md->data_lock);
  }
}

static status_t
kitty_command_send(mydev *md, uint16 command)
{
  status_t st;
  int i;
  char *data_buffer = md->buf;

  DPRINTF((ID "send_request(md=%p, command=%x)\n", md, command));
  DPRINTF((ID "send_request(md->dev=%p)\n", md->dev));

  st = acquire_sem(md->cmd_complete); // release_sem(cmd_complete)$B$O(Bcb_command_send()$B$NCf$G<B9T(B
  acquire_sem(md->data_lock);
  if ((st = usb->queue_request(md->dev, 
			       USB_REQTYPE_ENDPOINT_IN | USB_REQTYPE_VENDOR, // unit8 request_type
			       USB_REQUEST_SET_FEATURE, // uint8 request = 0x03
			       0x002C, // uint16 value
			       command, // uint16 index
			       0x0008, // uint16 length
			       md->buf, // void *data
			       8, // size_t data_len
			       cb_command_send, // usb_callback_func notify
			       md // void *cookie
			       )) != B_OK ) {
    DPRINTF((ID "queue_request() returns %d, failed!\n", st));
    return B_ERROR;
  }
  else {
    DPRINTF((ID "queue_request() returns %d, success!\n", st));
    DPRINTF((ID "data = 0x%02x 0x%02x 0x%02x 0x%02x 0x%02x 0x%02x 0x%02x 0x%02x \n",
	     *data_buffer++, *data_buffer++, *data_buffer++, *data_buffer++, 
	     *data_buffer++, *data_buffer++, *data_buffer++, *data_buffer));
    return B_OK;
  }
}

static status_t
init_device(mydev *new_dev, const usb_device *dev, const int32 num)
{
  status_t st;

  DPRINTF((ID "init_device(%p, %p, %x)\n", new_dev, dev, num));
  new_dev->dev = dev;
  new_dev->open = 0;
  new_dev->num = num;

  new_dev->lock = create_sem(1, ID "lock");
  new_dev->data_lock = create_sem(1, ID "data_lock");
  if (new_dev->lock < 0) {
    st = new_dev->lock;
    return st;
  }
  new_dev->cmd_complete = create_sem(1, ID "cmd_complete_lock");
  if (new_dev->cmd_complete < 0) {
    st = new_dev->cmd_complete;
    return st;
  }
  DPRINTF((ID "init_device() done.\n"));
  return B_OK;
}

static void
deinit_device(mydev* md)
{
  DPRINTF((ID "deinit_device(%p)\n", md));
  delete_area(md->buf_area);
  delete_sem(md->lock);
  delete_sem(md->data_lock);
  delete_sem(md->cmd_complete);
  DPRINTF((ID "deinit_device() done.\n"));
}

static mydev *
add_device(const usb_device *dev)
{
  mydev    *new_dev = NULL;
  int      i;
  status_t st;
  char     area_name[32];

  DPRINTF((ID "add_device(%p)\n", dev));
  if (device_count >= MAX_DEVICE_NUM) {
    return new_dev; /* return NULL */
  }

  acquire_sem(device_list_lock);
  if (new_dev = (mydev *)malloc(sizeof(mydev))) {
    init_device(new_dev, dev, device_count); /* new_dev$B9=B$BN$N=i4|2=(B */
    new_dev->next = device_list; /* device_list$B%j%9%H$K(Bnew_dev$B9=B$BN$r$D$J$0(B */
    device_list = new_dev; /* device_list$B%j%9%H$N@hF,$r(Bnew_dev$B$K99?7(B */
    DPRINTF((ID "device_count++: %d", device_count));
    device_count++; /* $B%0%m!<%P%kJQ?t(Bdevice_count$B$r%$%s%/%j%a%s%H(B */
    DPRINTF(("->%d\n", device_count));
  }

  /* $B%P%C%U%!:n@.(B */
  sprintf(area_name, "usb_kitty_buf%d", device_count-1);
  new_dev->buf_area = create_area(area_name, (void **)&new_dev->buf,
				  B_ANY_KERNEL_ADDRESS, B_PAGE_SIZE,
				  B_CONTIGUOUS, B_READ_AREA | B_WRITE_AREA);
  if (new_dev->buf_area < 0) {
    DPRINTF((ID "create_area() failed!\n", device_count));
  }
  
  release_sem(device_list_lock);
  DPRINTF((ID "add_device() done.\n"));
  return new_dev;
}

static status_t
device_added(const usb_device *dev, void **cookie)
{
  mydev *md;
  DPRINTF((ID "device_added(%p, %p)\n", dev, cookie));
  if ((md = add_device(dev)) != NULL) {
    *cookie = md;
    DPRINTF((ID "device_added() done.\n"));
    return B_OK;
  }
  else {
    DPRINTF((ID "add_device return NULL\n"));
  }
  DPRINTF((ID "device_added() done.\n"));
  return B_ERROR;
}

static void
remove_device(mydev* md)
{
  DPRINTF((ID "remove_device(%p)\n", md));

  deinit_device(md);
  free(md);
  if (md != device_list) {
    DPRINTF((ID "got a wired device in remove_device()\n"));
  }
  device_list = NULL;
  DPRINTF((ID "device_count--: %d", device_count));
  device_count--; /* $B%0%m!<%P%kJQ?t(Bdevice_count$B$r%G%/%j%a%s%H(B */
  DPRINTF(("->%d\n", device_count));
  DPRINTF((ID "remove_device() done.\n"));
}	

static status_t
device_removed(void *cookie)
{
  mydev *md = (mydev *)cookie;

  DPRINTF((ID "device_removed(%p)\n", cookie));
  acquire_sem(device_list_lock);
  remove_device(md);
  release_sem(device_list_lock);
  DPRINTF((ID "device_removed() done.\n"));
  return B_OK;
}

static usb_notify_hooks notify_hooks = {
  &device_added,
  &device_removed
};

static status_t
mydev_open(const char *name, uint32 flags, void **cookie)
{
  mydev *md;
  status_t st;
  int n;

  DPRINTF((ID "mydev_open(*name, flags, **cookie)\n"));
  DPRINTF((ID "mydev_open(%p, %x, %p)\n", name, flags, *cookie));
  n = atoi(name + strlen(base_name));
  //  DPRINTF((ID "mydev_open(\"%s\", %d,...)\n", name,flags));

  acquire_sem(device_list_lock);
  DPRINTF((ID "mydev_open() *device_list=%p)\n", device_list));
  for (md = device_list; md ; md = md->next) {
    /* device_list$B%j%9%H$N@hF,$+$i$9$Y$F$N%j%9%H$r8!:w(B */
    if (md->num == n) {
      md->open++;
      *cookie = md; /* $B$3$3$GF@$?(Bmydev$B9=B$BN(Bmd$B$r(Bopen()$B$G3MF@$7$?(Bcookie$B$H$7$F(B
		       $B0J9_$N(Bread()/write()$B$J$I$G(Bclose()$B$9$k$^$G;H$&(B */
      DPRINTF((ID "mydev_open() *md=%p)\n", md));
      break;
    }
  }
  release_sem(device_list_lock);

  /* $B%G%P%$%9$N=i4|2=$r<B9T(B */
  acquire_sem(md->lock);

  if ((md->conf = usb->get_nth_configuration(md->dev, 0)) == NULL) {
    DPRINTF((ID "get_nth_configuration failed!!\n"));
  }
  else {
    /* $B%G%P%$%9$N%3%s%U%#%.%e%l!<%7%g%s$N<hF@$K@.8y(B */
    DPRINTF((ID "get_nth_configuration success!!\n"));
    DPRINTF((ID "md->conf->descr = %p, md->conf->interface_count = %d md->conf->interface = %p\n",
	     md->conf->descr, md->conf->interface_count, md->conf->interface));
  }

  md->intf = md->conf->interface->active;
  if ((st = usb->set_alt_interface(md->dev, md->intf)) != B_OK) {
    DPRINTF((ID "set_alt_interface() failed! %x.\n", st));
    goto exit;
  }
  else {
    DPRINTF((ID "set_alt_interface() success! %x.\n", st));
  }
  if (usb->set_configuration(md->dev, md->conf) != B_OK) {
    DPRINTF((ID "set_configuration failed!\n"));
    st = B_ERROR;
    goto exit;
  }
  else {
    DPRINTF((ID "set_configuration success!\n"));
  }

  md->ept = md->conf->interface->active->endpoint;
  if (md->ept == NULL) {
    DPRINTF((ID "md->ept == NULL\n"));
    st = B_ERROR;
    goto exit;
  }
  else {
    DPRINTF((ID "md->ept->descr  = %p\n", md->ept->descr));
    DPRINTF((ID "md->ept->handle = %p\n", md->ept->handle));
  }

  st = acquire_sem(md->cmd_complete);
#if 1
  if ((st = usb->queue_request(md->dev, 
			       USB_REQTYPE_ENDPOINT_IN | USB_REQTYPE_VENDOR, // unit8 request_type
			       USB_REQUEST_SET_FEATURE, // uint8 request = 0x03
			       0x002C, // uint16 value
			       //			       B_KITTY_BUTTON_DISABLE, // uint16 index
			       B_KITTY_BUTTON_DISABLE, // uint16 index
			       0x0008, // uint16 length
			       (void *)&(data[0]), // void *data
			       8, // size_t data_len
			       cb_command_send, // usb_callback_func notify
			       md // void *cookie
			       )) != B_OK ) {
    DPRINTF((ID "queue_request() returns %d, failed!\n", st));
    goto exit;
  }
  else {
    DPRINTF((ID "queue_request(B_KITTY_BUTTON_DISABLE) returns %d, success!\n", st));
    DPRINTF((ID "data = 0x%02x 0x%02x 0x%02x 0x%02x 0x%02x 0x%02x 0x%02x 0x%02x \n",
	     data[0], data[1], data[2], data[3], data[4], data[5], data[6], data[7]));
    goto exit;
  }
#endif

exit:
  release_sem(md->lock);
  DPRINTF((ID "mydev_open() return %d\n", st));
  return st;
}

static status_t
mydev_close(void *cookie)
{
  mydev *md = (mydev *)cookie;
  DPRINTF((ID "mydev_close(%p)\n", cookie));
  DPRINTF((ID "mydev_close() done.\n"));
  return B_OK;
}

static status_t
mydev_free (void *cookie)
{
  status_t err = B_ERROR;
  mydev *md = (mydev *)cookie;
  
  DPRINTF((ID "mydev_free(%p)\n", cookie));
  acquire_sem(device_list_lock);
  if (--md->open == 0) {
    /* $B:G8e$N%/%i%$%"%s%H$,(Bclose()$B$7!$%G%P%$%9$,$D$J$,$C$F$$$J$$$J$i(B
       $B%G%P%$%9$r:o=|$9$k(B
    */
    // remove_device(md);
  }
  release_sem(device_list_lock);
  DPRINTF((ID "mydev_free() done.\n"));
 
  return err;
}

static status_t
mydev_control(void *cookie, uint32 msg, void *params, size_t size)
{
  mydev* md = (mydev *)cookie;
  status_t err = B_OK;
  uint16 command;

  DPRINTF((ID "mydev_control(cookie=%p, msg=%x, params=%p, size=%x)\n", cookie, msg, params, size));
  if (md == NULL) {
    DPRINTF((ID "Bad cookie.\n"));
    return ENODEV;
  }
  
  err = acquire_sem(md->lock);
  if (err != B_OK) {
    DPRINTF((ID "Couldn't acquire device lock.\n"));
    goto exit;
  }

  switch (msg) {
  case B_KITTY_FACE_LED_ON:
  case B_KITTY_FACE_LED_OFF:
  case B_KITTY_BUTTON_ENABLE:
  case B_KITTY_BUTTON_DISABLE:
  case B_KITTY_STEADY_POSITION:
  case B_KITTY_QUESTION:
  case B_KITTY_HAND_UP:
  case B_KITTY_HELLO:
  case B_KITTY_JAPANESE_HELLO:
  case B_KITTY_SEE_YA:
  case B_KITTY_LOTO:
  case B_KITTY_BINGO:
  case B_KITTY_FANTASTIC:
  case B_KITTY_LUCKY:
  case B_KITTY_OH_NO:
  case B_KITTY_TIMER:
  case B_KITTY_ITS_TIME:
  case B_KITTY_PI:
  case B_KITTY_PIRORI:
  case B_KITTY_PIPIPI:
  case B_KITTY_PIRORI2:
  case B_KITTY_PIPIPI2:
  case B_KITTY_PIRORI3:
    command = (uint16)msg;
    err = kitty_command_send(md, msg);
    if (err != B_OK) {
      DPRINTF((ID "issue_mydev_command() returned abnormal status, %ld\n", err));
    }
    else {
      DPRINTF((ID "issue_mydev_command(0x%0x) returned successfully\n", command));
    }		
  default:
    DPRINTF((ID "Invalid ioctl() opcode %ld.\n", msg));
    err = B_DEV_INVALID_IOCTL;
    break;
  }
  release_sem(md->lock);
  DPRINTF((ID "mydev_control() done.\n"));
exit:
  return err;
}

static status_t
mydev_read(void *cookie, off_t pos, void *buf, size_t *count)
{
  // $B$H$-$I$-%\%?%s%/%j%C%/$r$H$j$3$\$9$H$-$,$"$k!%(B
  mydev* md = (mydev *)cookie;
  status_t err = B_OK;
  status_t st;
  char *data_buffer = md->buf;

  //  DPRINTF((ID "mydev_read(%p, %x, %p, %p)\n", cookie, pos, buf, count));

  //  snooze(1000);
  err = acquire_sem(md->lock);
  if (err != B_OK) {
    DPRINTF((ID "Couldn't acquire device lock.\n"));
    goto exit;
  }

  err = acquire_sem(md->cmd_complete);
  err = acquire_sem(md->data_lock);    
  md->timestamp = system_time();

  if ((err = usb->queue_request(md->dev, 
			       USB_REQTYPE_ENDPOINT_IN | USB_REQTYPE_VENDOR, // unit8 request_type
			       USB_REQUEST_SET_FEATURE, // uint8 request = 0x03
			       0x002C, // uint16 value
			       0x0032, // uint16 index
			       0x0008, // uint16 length
			       md->buf, // void *data
			       8, // size_t data_len
			       cb_command_send, // usb_callback_func notify
			       md // void *cookie
			       )) != B_OK ) {
    DPRINTF((ID "queue_request() returns %d, failed!\n", err));
    goto exit;
  }
  else {
    st = acquire_sem(md->data_lock);    
#if 0
    DPRINTF((ID "data = 0x%02x 0x%02x 0x%02x 0x%02x 0x%02x 0x%02x 0x%02x 0x%02x \n",
	     *data_buffer++, *data_buffer++, *data_buffer++, *data_buffer++, 
	     *data_buffer++, *data_buffer++, *data_buffer++, *data_buffer));
#endif
    memcpy(buf , md->buf, 8);
    (char *)buf++;
    md->button = *((char *)buf);
    (char *)buf--;
#if 0
    DPRINTF((ID "md->button = 0x%x\n", md->button));
#endif
    // $B%G!<%?$O1&%/%j%C%/$G(B 0x34, $B:8%/%j%C%/$G(B 0x38, $BN>J}%/%j%C%/$G(B 0x3C$B$,JV$k(B
    // $BLs(B64msec(80$B2sJ,$N(Bqueue_request)$B$*$-$K(B0x70, 0x61$B$,B3$$$FJV$k(B
    // (0x70$B$N$_!$(B0x61$B$N$_$7$+JV$i$J$$$H$-$b$"$k(B)
    switch (md->button) {
    case 0x30:
      //DPRINTF((ID "button NOT pressed!\n"));
      err = B_OK;
      *count = 8;
      break;
    case 0x34: // $B1&%/%j%C%/(B
    case 0x38: // $B:8%/%j%C%/(B
    case 0x3C: // $BN>J}%/%j%C%/(B
      DPRINTF((ID "button pressed!, buf = %p, *buf = %x md->button = 0x%02x\n", buf, *((char *)buf), md->button));
      err = B_OK;
      *count = 8;
      break;
    case 0x00: // $B$O$8$a$K=P$k(B
    case 0x61: // $BLs(B80$B2sKh$K=P$k(B
    case 0x70: // $BLs(B80$B2sKh$K=P$k(B
      //     dprintf(ID "time = %x, button = %02x\n", md->timestamp, md->button);
      err = B_OK;
      *count = 8;
      break;
    case 0x71: // $B$^$l$K=P$k(B
    case 0x01: // $B$H$F$b$^$l$K=P$k(B
    case 0x08: // $B$H$F$b$^$l$K=P$k(B
    case 0x09: // $B$H$F$b$^$l$K=P$k(B
    case 0x26: // $B$H$F$b$^$l$K=P$k(B
    case 0x4C: // $B$H$F$b$^$l$K=P$k(B
    case 0x69: // $B%\%?%s$r%/%j%C%/$7$?$H$-$K$^$l$K=P$k(B
    default:
      //      DPRINTF((ID "button NOT pressed!\n"));
      DPRINTF((ID "read() return B_ERROR, button = %02x\n", md->button));
      err = B_ERROR;
      *count = 0;
      break;
    }
  }

exit:
  //  DPRINTF((ID "read() return %d\n", err));
  release_sem(md->data_lock);
  release_sem(md->lock);
  return err;
}

static status_t
mydev_write(void *cookie, off_t pos, const void *buf, size_t *count)
{
  DPRINTF((ID "device_write(%p, %x, %p, %p)\n", cookie, pos, buf, count));
  DPRINTF((ID "device_write() done.\n"));
  return B_ERROR;
}

usb_support_descriptor supported_device[] = 
{
  { 0, 0, 0, VENDOR_ID, PRODUCT_ID } 
  /* VENDOR_ID, PRODUCT_ID$B$K%^%C%A$9$k%G%P%$%9$rK\%I%i%$%P$G%5%]!<%H$9$k(B */
};

status_t
init_hardware(void) {
  /* USB$B%I%i%$%P$G$O%G%P%$%9$,B8:_$7$J$$>l9g$b$"$k$N$G!$$3$3$G$O2?$b$7$J$$(B */
  DPRINTF((ID "init_hardware()\n"));
  DPRINTF((ID "init_hardware() done.\n"));
  return B_OK;
}

status_t
init_driver(void) { /* init_driver()$B$O%I%i%$%P$,%m!<%I$5$l$?>l9g$K$O2?EY$b%3!<%k$5$l$k(B */
  DPRINTF((ID "init_driver()\n"));
  if (get_module(B_USB_MODULE_NAME, (module_info**)&usb) != B_OK) {
    /* USB$B%P%9%^%M!<%8%c$HDL?.$9$k:]$KI,MW$J%b%8%e!<%k>pJs$rJ]B8(B */
    return B_ERROR;
  }

  usb->register_driver(driver_name, supported_device, 1, NULL);
    /* $BK\%I%i%$%P$,%5%]!<%H$9$k%G%P%$%9(B(supported_device$B$K$h$j;XDj(B)$B$H(B
       $B%I%i%$%PL>$rEPO?(B */
  usb->install_notify(driver_name, &notify_hooks);
    /* USB$B%P%9%^%M!<%8%c$X%I%i%$%PL>$H!$A^H4;~$N%3!<%k%P%C%/4X?t$rEPO?(B */

  device_list_lock = create_sem(1, ID "device_list_lock");
  if (device_list_lock < 0) {
    put_module(B_USB_MODULE_NAME);
    return B_ERROR;
  }
  DPRINTF((ID "init_driver() done.\n"));
  return B_OK;
}

void
uninit_driver(void) {
  int i;

  DPRINTF((ID "uninit_driver()\n"));
  usb->uninstall_notify(driver_name);
  delete_sem(device_list_lock);
  put_module(B_USB_MODULE_NAME);

  /* $B%G%P%$%9L>J8;zNs$r(Bfree */
  if (device_names != NULL) {
    for(i=0;device_names[i];i++) {
      free((char *)device_names[i]);
    }
    free(device_names);
  }

  if (device_list != NULL) {
    DPRINTF((ID "Device %p still exists at teardown.\n", device_list));
  }
  DPRINTF((ID "uninit_driver() done.\n"));
}

const char **
publish_devices(void)
{
  mydev *md;
  int i;

  DPRINTF((ID "publish_device()\n"));
  /* $B%G%P%$%9L>J8;zNs$r0lC6(Bfree */
  if (device_names != NULL) {
    for(i=0;device_names[i];i++) {
      free((char *)device_names[i]);
      DPRINTF((ID "publish_device() ... free(%d)\n", i));
    }
    free(device_names);
  }

  acquire_sem(device_list_lock);
  DPRINTF((ID "publish_device() entering device_list_lock CS\n"));
  /* $B%G%P%$%9L>J8;zNs$r:FEY(Balloc */
  device_names = (char **)malloc(sizeof(char *)*(device_count + 1));
  if (device_names != NULL) {
    for (i=0, md=device_list ; md ; md = md->next) {
      if (device_names[i] = (char *)malloc(strlen(base_name)+4)) {
	sprintf(device_names[i], "%s%d", base_name, md->num);
	i++;
      }
    }
    device_names[i] = NULL;
  }
  else {
    DPRINTF((ID "publish_device() malloc() device_names failed!\n"));    
  }
  release_sem(device_list_lock);

  DPRINTF((ID "publish_device() done.\n"));
  return (const char **)device_names;
}

static device_hooks DeviceHooks = {
  mydev_open,
  mydev_close,
  mydev_free,
  mydev_control,
  mydev_read,
  mydev_write,
  NULL,
  NULL,
  NULL,
  NULL
};

device_hooks*
find_device(const char *name)
{
  DPRINTF((ID "find_device()\n"));
  DPRINTF((ID "find_device() done.\n"));
  return &DeviceHooks;
}
